import os
import shutil
from pathlib import Path
from collections.abc import Generator

from pytest import fixture


def get_all_genereated_files(path: Path) -> list[Path]:
    return sorted(
        [
            Path(root).joinpath(file)
            for (root, _, files) in os.walk(path, topdown=False)
            for file in files
            if not file.startswith(".")
        ]
    )


@fixture
def path() -> Generator[Path]:
    path = Path(os.environ["_SPY_RESULTS_DIR"])
    path.mkdir(exist_ok=True)
    yield path

    shutil.rmtree(path)



import json
import os
from pathlib import Path
from typing import Any
from data_surveillance.spy import spy
import pytest
import pandas as pd
import os
from .conftest import get_all_genereated_files


def test_spy_in_one_constrain(path: Path) -> None:
    # given
    value = pd.DataFrame(
        {
            "col_1": [1, 2, 3],
            "col_2": [11, 22, 33],
        },
    )
    # expected
    expected_value = pd.DataFrame(
        {
            "col_2": [11, 22, 33],
        },
    )

    @spy(
        constrains=[
            (pd.DataFrame, lambda x: x.drop(columns=["col_1"]), "in"),
        ]
    )
    def calc(value):
        pass

    # when
    calc(value=value.copy())

    # then
    files = get_all_genereated_files(path=path)
    pd.testing.assert_frame_equal(expected_value, pd.read_csv(files[0], index_col=0))


def test_spy_in_two_constrains(path: Path) -> None:
    # given
    value = pd.DataFrame(
        {
            "col_1": [1, 2, 3],
            "col_2": [11, 22, 33],
            "col_3": [11, 22, 33],
        },
    )
    # expected
    expected_value = pd.DataFrame(
        {
            "col_2": [11, 22, 33],
        },
    )

    @spy(
        constrains=[
            (pd.DataFrame, lambda x: x.drop(columns=["col_1"]), "in"),
            (pd.DataFrame, lambda x: x.drop(columns=["col_3"]), "in"),
        ]
    )
    def calc(value):
        pass

    # when
    calc(value=value.copy())

    # then
    files = get_all_genereated_files(path=path)
    pd.testing.assert_frame_equal(expected_value, pd.read_csv(files[0], index_col=0))


def test_spy_in_out_constrains(path: Path) -> None:
    # given
    value = pd.DataFrame(
        {
            "col_1": [1, 2, 3],
            "col_2": [11, 22, 33],
            "col_3": [11, 22, 33],
        },
    )
    # expected
    expected_value = pd.DataFrame(
        {
            "col_2": [22, 44, 66],
        },
    )

    def advance(v: pd.DataFrame) -> pd.DataFrame:
        v["col_2"] = [22, 44, 66]
        return v

    @spy(
        constrains=[
            (pd.DataFrame, lambda x: x.drop(columns=["col_1"]), "both"),
            (pd.DataFrame, lambda x: x.drop(columns=["col_3"]), "both"),
            (pd.DataFrame, advance, "out"),
        ]
    )
    def calc(value):
        return value

    # when
    calc(value=value.copy())

    # then
    files = get_all_genereated_files(path=path)

    pd.testing.assert_frame_equal(expected_value, pd.read_csv(files[1], index_col=0))


def test_spy_in_out_constrains_v2(path: Path) -> None:
    # given
    value = pd.DataFrame(
        {
            "col_1": [1, 2, 3],
            "col_2": [11, 22, 33],
            "col_3": [11, 22, 33],
        },
    )
    # expected
    expected_value = pd.DataFrame(
        {
            "col_1": [1, 2, 3],
            "col_2": [22, 44, 66],
            "col_3": [11, 22, 33],
        },
    )

    def advance(v: pd.DataFrame) -> pd.DataFrame:
        v = v.copy()
        v["col_2"] = [22, 44, 66]
        return v

    @spy(
        constrains=[
            (pd.DataFrame, lambda x: x.drop(columns=["col_1"]), "in"),
            (pd.DataFrame, lambda x: x.drop(columns=["col_3"]), "in"),
            (pd.DataFrame, advance, "out"),
        ]
    )
    def calc(value):
        return value

    # when
    calc(value=value.copy())

    # then
    files = get_all_genereated_files(path=path)

    pd.testing.assert_frame_equal(expected_value, pd.read_csv(files[1], index_col=0))



import os
from pathlib import Path
from random import randint
import logging
from logging import StreamHandler

if _path := os.environ.get("_SPY_RESULTS_DIR"):
    SPY_RESULTS_DIR = Path(_path)
else:
    SPY_RESULTS_DIR = Path.cwd() / f".{randint(1000, 9999)}"


HASH_OUTPUT_FILE_NAME: str = ".hashed.json"


logger = logging.getLogger(__name__)
logger.addHandler(StreamHandler())
logger.setLevel(logging.INFO)



import json
import os
from pathlib import Path
import pickle
from typing import Any, Callable, Iterable, Literal, Type, Type

import pandas as pd
from .bootstrap import SPY_RESULTS_DIR, logger


def _save(
    root_folder: Path,
    step: int,
    value: Any,
    key: None | str = None,
    opt_type: Literal["in", "out", ""] = "in",
    file_name: str = "",
) -> None:
    _step = f"0{step}" if step < 10 else str(step)

    file_name += f"{opt_type}__{_step}__type={type(value).__name__}"

    if key:
        file_name = f"{file_name}__key={key}"

    if isinstance(value, pd.Series):
        _value = value.copy()
        file_name += ".json"
        _value = sorted(_value.to_list())
        with root_folder.joinpath(file_name).open("w") as f:
            f.writelines(json.dumps(_value))
    elif isinstance(value, pd.DataFrame):
        _value = value.copy()

        _value = _value.sort_index(axis=1)  # sor ccols
        _value = _value.sort_values(by=[_ for _ in _value.columns])
        _value = _value.reset_index(drop=True)

        file_name += ".csv"
        _value.to_csv(root_folder.joinpath(file_name))
    elif (
        isinstance(value, (list, tuple))
        and len(value) > 0
        and any(isinstance(_, (pd.DataFrame, pd.Series)) for _ in value)
    ):
        for i, element in enumerate(value):
            _save(root_folder, i, element, None, "", file_name)
    else:
        try:
            to_file = json.dumps(value)
            file_name += ".json"
        except TypeError as ex:
            logger.debug(ex.args[0])
            to_file = str(pickle.dumps(value))
            file_name += ".pickle"

        with root_folder.joinpath(file_name).open("w") as f:
            f.writelines(to_file)


def _apply_constrains(
    constrains: list[tuple[type, Callable, Literal["in", "out", "both"]]], v: Any
) -> Any:
    for c_type, v_func, _ in constrains:
        if isinstance(v, c_type):
            v = v_func(v)
    return v


def spy(
    constrains: list[tuple[type, Callable, Literal["in", "out", "both"]]] | None = None,
) -> Callable:
    path = SPY_RESULTS_DIR

    if not path.exists():
        path.mkdir(exist_ok=True)

    def inner(func) -> Callable:
        def wrapper(*arg, **kwargs):  # noqa: ANN002, ANN003, ANN202
            _constrains = () if constrains is None else constrains
            counter_file = path / ".counter"
            if counter_file.exists():
                step = counter_file.read_text()
                step = "0" if step == "" else str(int(step) + 1)
                counter_file.write_text(step)
            else:
                step = "0"
                counter_file.write_text(step)

            _step = f"0{step}" if int(step) < 10 else str(step)

            step_folder = path.joinpath(f"step_{_step}__{func.__qualname__}")
            step_folder.mkdir(exist_ok=True)

            small_step = 0
            _in_constrains = list(filter(lambda x: x[2] in ["in", "both"], _constrains))
            for v in arg:
                _v = _apply_constrains(_in_constrains, v) if _in_constrains else v

                _save(step_folder, small_step, _v, None, "in")
                small_step += 1

            for k, v in kwargs.items():
                _v = _apply_constrains(_in_constrains, v) if _in_constrains else v
                _save(step_folder, small_step, _v, k, "in")
                small_step += 1

            r = func(*arg, **kwargs)
            if r is None:
                return None

            _out_constrains = list(
                filter(lambda x: x[2] in ["out", "both"], _constrains)
            )

            if isinstance(r, (list, tuple)):
                for v_r in r:
                    _v_r = (
                        _apply_constrains(_out_constrains, v_r)
                        if _out_constrains
                        else v_r
                    )

                    _save(step_folder, small_step, _v_r, None, "out")
                    small_step += 1
            else:
                _v_r = _apply_constrains(_out_constrains, r) if _out_constrains else r
                _save(step_folder, small_step, _v_r, None, "out")
            return r

        return wrapper

    return inner


# # TODO: catch input output X
# # TODO: do numbers {00, 01} not {0,1} X
# # TODO: add other python objects X
# # TODO: Add hash comaprison X

# # TODO: add script which adding to every function my decorator:) -> orders with @staticfunction
